function [Clustership,  ClusterCount,  O_prototypesA, O_prototypesAKeyGenes, O_prototypesAKeyGenesNames, keyGenesNames, O_prototypesC, C_prototypes] = SAModKprototypes()
%=====================================================================================================
% Program:              SA-ModkPrototypes.m
% Programmer:           Pierre R. Bushel
% Programming Language: Matlab
% Date:                 July 16th, 2005
%
% Description: Matlab program to cluster gene expression data and clinical chemistry evaluations as 
% numeric data with categorical values from histopathology observations using an approach similar to Zhexue Huang (1997) 
% k-prototype algorithm with objective function optimization based on the Kirkpatrick, Gelatt Jr., and Vecchi (1983)
% simulated annealing procedure with Metropolis Monte Carlo sampling (simulations) and a simple exponential cooling 
% schedule (Salamon, Sibani and Frost, 2002). 
%
% Theory: For the k-prototype algorithm, the total cost for assigning object X to cluster_l when X has categorical attributes
% is minimized if and only if the frequency of categorical attribute A_j = the value of the categorical attribute for the prototype 
% of cluster_l (q_l) given a set of categorical objects (X)) is greater than or equal to the frequency of categorical attribute 
% A_j = the value of the kth category in attribute A_j (C_k_j) given a set of categorical objects (X)) for q_j notequal to c_k_j 
% for all categorical attributes.

% The simulated annealing algorithm guarantees that after a sufficient number of iterations, the system satisfies the Boltzmann
% distribution at a given temperature.  Therefore, if the temputature of the system approaches zero slowly enough, the system will 
% avoid a local minimum in search of the global minimum of the surface of the objective function.


% Data Types and Format: Gene expression data needs to be log transformed and formatted (short and wide) in a tab delimited text file with 
% array observations as row values and gene, clinical chemistry and histopathology variables as column values.  The first row is the column header, 
% the second row is an integer denoting the data type (1 = gene expression, 2 = clinical chemistry measurement, 3 = histopathology observation).  
% The first column contains sample names and the first row contains the variable names.  The order of the data in the file should be from data 
% type 3 to 2, to 1 and be within individual groups or blocks.    

% Limitations: Only one categorical feature value per observation is permitted.  A feature can exist as only one type of data.  
% The application is optimized for clustering the samples and identifying phenotypic prototypes from the groups of them, not of the genes.  
% The application is not guaranteed to find the optimal solution for the clustering of the samples, just the assignment of the samples to clusters 
% according to the reduction of an objective function close to the global minimum.  The weights of the components of the objective function must 
% sum up to 1.0 and either a significance level of 0.05 or 0.01 can be used for identifying vetted genes.  The program has been tested in Windows 2000 and XP environments. 

% Requirements: SA-ModkPrototypes is a set of Matlab functions and scripts tested in Matlab version 6.5.X.X R13.X for Windows (98/NT4/2000).  
% You may encounter problems in other operating systems, platforms and/or other Matlab versions.  The applications require the Matlab Statistics Toolbox Version 4.0, 
% the Resampling Stats Toolbox Version 1.0 by Daniel T. Kaplan (Department of Mathematics and Computer Science, Macalester College, St. Paul,Minnesota, USA), 
% the adjusted Rand Index function by Tijl De Bie(February 2003), the Matlab loadcell.m function to load mixed type data and the cell2csv.m function to convert cell arrays to 
% comma separate value formated files, both available at the Matlab Central File Exchange (File ID 1965 and 7601 respectively).  
% Be sure to set the path of the Toolboxes in Matlab before running the application.
% Comment out line 1 (function....) in order to run this application in the Matlab native environment.
%
%
%=====================================================================================================
%clear all;
warning off;
close all;
%echo off;
clc;

%================================================================
% Splash Screen 
%================================================================
disp('*********************************************************');
disp('*         SA-Modk-Prototypes Clustering Algorithm       *');
disp('*         Version 1.0                                   *');
disp('*         Developed by Pierre R. Bushel                 *');
disp('*         at NIEHS                                      *');
disp('*********************************************************');
%================================================================
% Global Variables
%================================================================

%================================================================
% Keyboard input 
%================================================================
theDataFile = input('Enter the path to and the name of the data file: ','s'); % Data file path and name string
doScale = input('Do you want to scale the data (1=yes, 0=no): '); % Sacle the data?
%theTransAns = input('Log base 2 transform gene expression data (1=yes, 0 = no): '); % 
theClustNum = input('Enter the number of k clusters desired: '); % Number of clusters used for partition 
%theSeed = input('Enter the seed: '); %  seed for random number generator
theInitialTemp = input('Enter the initial temperature for the cooling schedule: '); %  T_zero set = 1.0
theFinalTemp = input('Enter the final temperature for the cooling schedule: '); %  T_final set = 0.001
theCoolingRate = input('Enter the cooling rate factor for the schedule: '); %  cooling rate fator set = 0.01
theStepsNum = input('Enter the number of steps for the Metropolis algorithm (100 - 10000): '); %  N_steps
%theStepsMax = input('Enter the maximum number of steps for the cooling schedule (10 - 100): '); %  N_max
%theStepsStop = input('Enter the number of steps with no energy change for stopping (3 - 5): '); %  N_steps_stop

%theClustCycleNum = input('Enter the number of cycles for clustering: '); % Number of cycles for clustering
theAlphaWeight = input('Enter a value for the Array weighting factor: '); % k-prototype alpha weight
theBetaWeight = input('Enter a value for the Clinical Chem weighting factor: '); % k-prototype beta weight
theGammaWeight = input('Enter a value for the Histopath weighting factor: '); % k-prototype gamma weight
setSigLevel = input('What significance level do you want to identify genes at (0.05 or 0.01): '); % set significance level?
if setSigLevel ~= 0.05
    if setSigLevel ~= 0.01
        setSigLevel = 0.05;
        disp('Significance level automatically set at 0.05');
    end
end
if setSigLevel == 0.05
    theAlphaLevel = 3.8415;
elseif setSigLevel == 0.01
    theAlphaLevel = 6.6349;
end
doRand = input('Do you want to comupte the Adjusted Rand index (1=yes, 0=no): '); % run Adjusted Rand?

if doRand ==1
    true_class = input('Enter the true classification of the samples(open\close brackets [] with spaces between numbers): '); % classes
    true_class_narrow = load(true_class_file);
    true_class = true_class_narrow';
end
%[theData, theVariables, theArrays] = tblread(theDataFile,'\t'); % 
%tdfread('.\Data\example.txt','\t'); %
%[theCellData,b]=loadcell(theDataFile,[',' char(9)],'"','single-string');
[theCellData,b]=loadcell(theDataFile,[char(9)],'"','single-string');
%warning off MATLAB:divideByZero
%[theCellData,b]=loadcell('.\data\simulated_data_trimmed_04.txt',[',' char(9)],'"','single-string');
[theCellData_n, theCellData_p] = size(theCellData);

%theClustNum = 5;
%theSeed = 123;
%theGammaWeight = 0;
%theInitialTemp = 1.0;
%theStepsNum = 100;
%theCoolingRate = .10;
%theCellData_n = theCellData_n - 1;
%NumData = [];
%CatData = [];

for i = 3:theCellData_n
    daCounter = 0;
    for j = 2:theCellData_p
        if str2num(char(theCellData(2,j))) == 1 
            daCounter = daCounter +1;
			if strcmp(theCellData(i,j),'.') == 1 
                NumDataArray(i-2,daCounter) = 0;
            else
                NumDataArray(i-2,daCounter) = str2num(char(theCellData(i,j)));
            end
        end
    end
end
for i = 3:theCellData_n
    daCounter = 0;
    for j = 2:theCellData_p
        if str2num(char(theCellData(2,j))) == 2
            daCounter = daCounter +1;
            if strcmp(theCellData(i,j),'.') == 1 
                 NumDataChem(i-2,daCounter) = 0;
            else
                NumDataChem(i-2,daCounter) = str2num(char(theCellData(i,j)));
            end
        end
    end
end
for i = 3:theCellData_n
    daCounter = 0;
    for j = 2:theCellData_p
        if str2num(char(theCellData(2,j))) == 3 
            daCounter = daCounter +1;
            CatData{i-2,daCounter} = char(theCellData(i,j));
        end 
    end
end        
theRangeArray= range(NumDataArray);
theRangeChem= range(NumDataChem); 
[theNumDataArray_n, theNumDataArray_p] = size(NumDataArray);
[theNumDataChem_n, theNumDataChem_p] = size(NumDataChem);
[theCatData_n, theCatData_p] = size(CatData);


%================================================================
% Scale the data 
%================================================================
if doScale ==1
	NumDataArrayCopy = NumDataArray;
	NumDataChemCopy = NumDataChem;
	for i = 1:(theCellData_n-2)
        NumDataArray_prescaled(i,:) = NumDataArray(i,:)./theRangeArray;
	end
	
	NumDataArray_prescaledSum = sum(NumDataArray_prescaled);
	
	for i = 1:(theCellData_n-2)
        NumDataArray(i,:) = NumDataArray_prescaled(i,:)./NumDataArray_prescaledSum;
	end
	%NumDataArraySum = sum(NumDataArray_scaled);
	for i = 1:(theCellData_n-2)
        NumDataChem_prescaled(i,:) = NumDataChem(i,:)./theRangeChem;
	end
	
	NumDataChem_prescaledSum = sum(NumDataChem_prescaled);
	
	for i = 1:(theCellData_n-2)
        NumDataChem(i,:) = NumDataChem_prescaled(i,:)./NumDataChem_prescaledSum;
	end
	%NumDataArraySum = sum(NumDataChem_scaled);
end

[theData, theVariables, theArrays] = tblread(theDataFile,'\t'); %
arrayNamesRow = cellstr(theArrays);
variableNamesRow = cellstr(theVariables);
[theSampleNum, crap] = size(arrayNamesRow);
%if theClustNum >= theSampleNum
    %fprintf('K not less than the number of samples in the data set,  Exiting Matlab');
    %disp('K not less than the number of samples in the data set,  Exiting Matlab');
    %exit
    %end
[theData_n, theData_p] = size(theData);
% theTrainSetSelect = input('Enter the columns of the data file representing the training set: '); % Columns of file with training data
%================================================================
% Preprocessing of data 
%================================================================

%if theTransAns == 1
    % Log transformation 
    %thePreProcessedData = log2(theData);
%else
    %thePreProcessedData = theData;
%end
    % Normalization across arrays
%theNormData = [];
%for i = 1:theData_n
%   theNormData(i,:) = (theData(i,:) - mean(theData(i,:)))/std(theData(i,:));
%end
    % Normalization within array
    % Filtering 

%================================================================
% Differentially expressed gene selection methods
%================================================================
    % Standard deviation approach assumes approximately normal distribution and 
    % low noise variance in proportion to ratio expression level distibution
theGeneTracker = [];
theClinChemTracker = [];
theHistoPathTracker = [];
theNewData = [];




counterA = 0;
counterB = 0;
counterC = 0;
for i =  2:theCellData_p
    if str2num(char(theCellData(2,i))) == 1
        counterA = counterA + 1;
        theGeneTracker(counterA) = i;
    elseif str2num(char(theCellData(2,i))) == 2
        counterB = counterB + 1;
        theClinChemTracker(counterB) = i;
    else
        counterC = counterC + 1;
        theHistoPathTracker(counterC) = i;
    end
end
T0 = clock;
%General_Silhouettes = [];
%General_Silhouettes(1) = 0;

[B,arrayIndex,J] = unique(arrayNamesRow(2:theSampleNum,1));

%seed(theSeed);

%================================================================
% Zhexue Huang (1997) K-Prototype cost function for mixed data types
%================================================================
%================================================================
% Initialization % Random selection of prototypes to initialize the algorithm
%================================================================
        
randomized = resamp(theClustNum,arrayIndex,1);

%Assign each object to a cluster whose prototype is nearest to it according
%to the mixed similarity measure (squared Euclidean distance for numeric
%values, weighted simple matching for categorical values)

%For each object_i
        %assign cluster membership = arg min distance(object_i, Cluster_j)
%End
ClusterCount = zeros(size(randomized));
sumInClusterArray = zeros(theClustNum,theNumDataArray_p);
sumInClusterChem = zeros(theClustNum,theNumDataChem_p);
O_prototypesArray = zeros(theClustNum,theNumDataArray_p);
O_prototypesChem = zeros(theClustNum,theNumDataChem_p);
for k = 1:theClustNum
    for j = 1:theNumDataArray_p
        O_prototypesArray(k,j) = NumDataArray(randomized(k),j);
    end
end
for k = 1:theClustNum
    for j = 1:theNumDataChem_p
        O_prototypesChem(k,j) = NumDataChem(randomized(k),j);
    end
end
for k = 1:theClustNum
    for j = 1:theCatData_p
        C_prototypes(k,j) = CatData(randomized(k),j);
    end
end
for i = 1:theCatData_p
    XXXX = unique(CatData(:,i));
    [theCat_n, theCat_p] = size(XXXX);
    theCatArray(i) = theCat_n;
end
theCat_n_Max = max(theCatArray);
frequencyInCluster = zeros(theClustNum, theCatData_p, theCat_n_Max);

for i = 1:theCellData_n-2
    minNumDistanceArray = distance(NumDataArray(i,:), O_prototypesArray(1,:), (theNumDataArray_p));
    minNumDistanceChem = distance(NumDataChem(i,:), O_prototypesChem(1,:), (theNumDataChem_p));
    minCatDistance = sigma(CatData(i,:), C_prototypes(1,:), theCatData_p);
    minDistance = theAlphaWeight*minNumDistanceArray + theBetaWeight*minNumDistanceChem + theGammaWeight*minCatDistance;
    Cluster = 1;
    for k = 1:theClustNum
        NumDistanceArray = distance(NumDataArray(i,:), O_prototypesArray(k,:), (theNumDataArray_p));
        NumDistanceChem = distance(NumDataChem(i,:), O_prototypesChem(k,:), (theNumDataChem_p));
	    CatDistance = sigma(CatData(i,:), C_prototypes(k,:), theCatData_p);
        Distance = theAlphaWeight*NumDistanceArray + theBetaWeight*NumDistanceChem + theGammaWeight*CatDistance;
        if Distance < minDistance
            minDistance = Distance;
            Cluster = k;
        end
	end
    Clustership(i) = Cluster;
    ClusterCount(Cluster) = ClusterCount(Cluster)+1; 
end

%================================================================
% Simulated Annealing Recursion % Reallocation of object cluster
% membership\Minimize cost function
%================================================================       

%While the number of moves has not chaged or the cost function has not changed in X number of cycles
        %For each cluster
                %recompute cluster_prototype_j
        %End
        %For each object_i
                %assign cluster membership = arg min distance(object_i, Cluster_j)
        %End
        %Compute and track value of cost function (E_t)
        %================================================================
		% Termination % End the program
		%================================================================      
        %If E_t = E_t-1 = E_t-2
                %Terminate 
        %End
        
%End

%================================================================
% 
%================================================================ 
 Distance_new = zeros(theNumDataArray_n, theClustNum);
 for i = 1:theCellData_n-2
     for k = 1:theClustNum
     
     
    %minNumDistance = distance(NumData(i,:), O_prototypes(1,:), theNumData_p);
    %minCatDistance = sigma(CatData(i,:), C_prototypes(1,:), theCatData_p);
    %minDistance = minNumDistance + theGammaWeight*minCatDistance;
    %Cluster = 1;
   
        if Clustership(i) == k
            NumDistanceArray = distance(NumDataArray(i,:), O_prototypesArray(k,:), (theNumDataArray_p));
            NumDistanceChem = distance(NumDataChem(i,:), O_prototypesChem(k,:), (theNumDataChem_p));
            Distance_new(i,k) = theAlphaWeight*NumDistanceArray + theBetaWeight*NumDistanceChem + theGammaWeight*CatDistance;
        end
    end
 end
   
preCostFunction_P = sum(Distance_new);
costFunction_O = sum(preCostFunction_P);

theInitialTemp
p0 = O_prototypesArray;
p1 = O_prototypesChem;
p2 = C_prototypes;
CC = ClusterCount;
SICA = sumInClusterArray;
SICC = sumInClusterChem;
FIC = frequencyInCluster;
while theInitialTemp >= theFinalTemp
   [Clustership, costFunction_O, p0, p1, p2, CC, SICA, SICC, FIC] = metropolis(NumDataArray, NumDataChem, CatData, Clustership, theInitialTemp, costFunction_O,  theStepsNum, CC, SICA, SICC, p0, p1, p2, FIC, theClustNum, theAlphaWeight, theBetaWeight, theGammaWeight);
   theInitialTemp = next_t(theInitialTemp, theCoolingRate);
   Temperture = theInitialTemp
end
ClusterCount = CC;
O_prototypesA = p0;
O_prototypesC = p1;
C_prototypes = p2;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Cluster Silhouettes  computation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Cluster_Silhouettes = zeros(1,theClustNum);
NumData = concat(NumDataChem, NumDataArray);
mySilhouettes = silhouette(NumData, Clustership);
for j = 1:theClustNum
    for i = 1:theCellData_n-2
        if Clustership(i) == j
            Cluster_Silhouettes(j) = mySilhouettes(i) + Cluster_Silhouettes(j);
        end
    end
    Cluster_Silhouettes(j) = Cluster_Silhouettes(j)/ClusterCount(j);
end
%cdtable = squareform(pdist(O_prototypes));
%cdtable_ave = (sum(sum(cdtable)))/(find_k*find_k-find_k);
General_Silhouettes = (sum(Cluster_Silhouettes)/theClustNum);%/(exp(1/cdtable_ave));

if doRand == 1    
    myx = adjrand(Clustership,true_class);
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Vet the genes using a chi square goodness of fit test
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
diffCounter = 1;
for i = 1:(theClustNum - 1)
    for j = (i+1):theClustNum
        for k = 1:theNumDataArray_p
            myAbsDifferenceArray(diffCounter) = abs(O_prototypesA(i,k)-O_prototypesA(j,k));
            diffCounter = diffCounter +1;
        end
    end
end
expectedDifference = mean(myAbsDifferenceArray);
for i = 1:(theNumDataArray_p*(Combin((theClustNum),2)))
    myChiSquareCriticalValues(i) = 2*(((myAbsDifferenceArray(i)-expectedDifference)^2)/expectedDifference);
end
myChiSquareCriticalValues_tran = myChiSquareCriticalValues';
lastCounter = 1;
lastlastcounter = 1;
significantGenes = [];
for i = 1:(theClustNum - 1)
    for j = (i+1):theClustNum
        for k = 1:theNumDataArray_p
            if myChiSquareCriticalValues(lastlastcounter) >= theAlphaLevel
                significantGenes(lastCounter, 1) = k;
                significantGenes(lastCounter, 2) = i;
                significantGenes(lastCounter, 3) = j;
                lastCounter = lastCounter +1;
            end
            lastlastcounter = lastlastcounter +1;
        end
    end
end
O_prototypesAKeyGenes = [];
keyGenesNames = {};
keyGenesNamesUnsorted = {};
if (isempty(significantGenes)) ~= 1
    [significantUniqueGenes, signIndex_I, signIndex_J] = unique(significantGenes(:,1));
    [thesignIndex_n, thesignIndex_p] = size(signIndex_I);

    for i=1:thesignIndex_n
        keyGenes(i,:) = significantGenes(signIndex_I(i),:);
    end
    %keyGenesSorted = sortrows(keyGenes,[2 3]);
    keyGenesSorted = sortrows(keyGenes,1);
    for i = 1:thesignIndex_n
       keyGenesNames(i,1) = (variableNamesRow(keyGenesSorted(i,1) + (theCatData_p + theNumDataChem_p)));
        keyGenesNamesUnsorted(i,1) = (variableNamesRow(keyGenes(i,1) + (theCatData_p + theNumDataChem_p)));
    end

    for j=1:theClustNum
        for i=1:thesignIndex_n
            O_prototypesAKeyGenes(j,i) = O_prototypesA(j,keyGenes(i,1));
        end
    end
    O_prototypesAKeyGenesNames = {};
    [theKeyGenes_n, theKeyGenes_p] = size(keyGenesNamesUnsorted);
    for i = 1:(theClustNum + 1)
        for j = 1:(theKeyGenes_n +1)
            if (i == 1) && (j ==1)
                O_prototypesAKeyGenesNames(i,j) = cellstr('Prototype Number');
            elseif (i == 1) && (j ~= 1)
                O_prototypesAKeyGenesNames(i,j) = keyGenesNamesUnsorted(j-1);
            elseif (i ~= 1) && (j == 1)
                O_prototypesAKeyGenesNames(i,j) = num2cell((i-1));
            else
                O_prototypesAKeyGenesNames(i,j) = num2cell(O_prototypesAKeyGenes(i-1,j-1));
            end
        end
    end
     
    save array_prototypes_key_genes.txt O_prototypesAKeyGenesNames -ascii -tabs;
    cell2csv('key_gene_names.txt',keyGenesNames,'\t');
else
    disp('No significant genes detected');
    %save array_prototypes_key_genes.txt O_prototypesAKeyGenes -ascii -tabs;
end
save cluster_assignment.txt Clustership -ascii -tabs;
save array_prototypes.txt O_prototypesA -ascii -tabs;
cell2csv('histopath_prototypes.txt',C_prototypes,'\t');
%cell2csv('key_gene_names.txt',keyGenesNames,'\t');
save clinchem_prototypes.txt O_prototypesC -ascii -tabs;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% End of program
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
T1 = clock;
timeElapsed = etime(T1,T0)/60;
if timeElapsed < 60
    timeElapsed = timeElapsed
else
    timeElapsed = timeElapsed/60
end


